/*****************************************************************************
 *
 * Atmel Corporation
 *
 * ̧		: HEXParser.cpp
 * ߲	: Dev-C++ 4.9.8.0 - http://bloodshed.net/dev/
 * 		: $Revision: 1163 $
 * t		: $Date: 2006N82 15:38:16 +0200 $
* XV	: $Author: ohlia $
*
* xҰ	: avr@atmel.com
*
* ړI	: Win32
*
* pLq	: AVR911 - AVRJ݊
*
* 		: ȒPIntel HEXٌ̧`ǂݏ
 *
 ****************************************************************************/

#include "HEXParser.hpp"

/* HEXںނǗ邽߂̓\ */
struct HEXRecord							// Intel HEX̧ ں
{
	unsigned char length;					// ް ޲Đłںޒ
	unsigned long offset;					// ̾ ڽ
	unsigned char type;						// ںތ`
	unsigned char * data;					// CӑIް ޲
};

void HEXFile::writeRecord( ofstream & f, HEXRecord * recp )
{
	unsigned char checksum;
	long recordPos;							// ں ް̈̈ʒu

	/* ьvZ */
	checksum = recp->length;
	checksum += (unsigned char) ((recp->offset >> 8) & 0xff);
	checksum += (unsigned char) (recp->offset & 0xff);
	checksum += recp->type;

	/* ں ͯޏ */
	f.fill('0');
	f << ":" << hex
		<< setw(2) << (long) recp->length
		<< setw(4) << (long) recp->offset
		<< setw(2) << (long) recp->type;

	/* ް ޲ď */
	for( recordPos = 0; recordPos < recp->length; recordPos++ )
	{
		checksum += recp->data[ recordPos ];	// XȂьvZ
		f << hex << setw(2) << (long) recp->data[ recordPos ];
	}

	/* я */
	checksum = 0 - checksum;				// ŏIя
	f << setw(2) << (long) checksum << endl;

	/* ُ팟 */
	if( !f.good() )
		throw new ErrorMsg( "Error writing HEX record to file!" );
}

void HEXFile::parseRecord( const string & hexLine, HEXRecord * recp )
{
	unsigned char checksum;
	long recordPos;							// ں ް̈ʒu

	if( hexLine.size() < 11 )				// Œ11ŁA
		throw new ErrorMsg( "Wrong HEX file format, missing fields! "
							"Line from file was: (" + hexLine + ")." );

	/* sɑ΂ٌ̍ */
	if( hexLine[0] != ':' )					// ':'ŊJn
		throw new ErrorMsg( "Wrong HEX file format, does not start with colon! "
							"Line from file was: (" + hexLine + ")." );

	/* A̾āA` */
	recp->length = Util.convertHex( hexLine.substr( 1, 2 ) );
	recp->offset = Util.convertHex( hexLine.substr( 3, 4 ) );
	recp->type = Util.convertHex( hexLine.substr( 7, 2 ) );

	/* ںނׂm */
	if( hexLine.size() < (11+recp->length*2) )
		throw new ErrorMsg( "Wrong HEX file format, missing fields! "
				"Line from file was: (" + hexLine + ")." );

	/* я */
	checksum = recp->length;
	checksum += (unsigned char) ((recp->offset >> 8) & 0xff);
	checksum += (unsigned char) (recp->offset & 0xff);
	checksum += recp->type;

	/* ް̈ */
	if( recp->length )
	{
		recp->data = new unsigned char[ recp->length ];

		/* ںންǂݍ */
		for( recordPos = 0; recordPos < recp->length; recordPos++ )
		{
			recp->data[ recordPos ] = Util.convertHex( hexLine.substr( 9 + recordPos*2, 2 ) );
			checksum += recp->data[ recordPos ];
		}
	}

	/* ? */
	checksum += Util.convertHex( hexLine.substr( 9 + recp->length*2, 2 ) );
	if( checksum != 0 )
	{
		throw new ErrorMsg( "Wrong checksum for HEX record! "
				"Line from file was: (" + hexLine + ")." );
	}
}

/* ݽ׸ */
HEXFile::HEXFile( long buffersize, long value )
{
	if( buffersize <= 0 )
		throw new ErrorMsg( "Cannot have zero-size HEX buffer!" );

	data = new unsigned char[ buffersize ];

	if( !data )
		throw new ErrorMsg( "Memory allocation failed for HEX-line-buffer!" );

	size = buffersize;

	clearAll( value );
}

/* ޽׸ */
HEXFile::~HEXFile()
{
	if( data ) delete data;
}

void HEXFile::readFile( const string & _filename )
{
    ifstream f;
	string hexLine;							// HEX̧ق1s܂
	HEXRecord rec;							// ꎞں

	long baseAddress;						// gڽw`ԗpڽ
	long dataPos;							// ںޓްʒu

	/* ̧قJƂ݂ */
	f.open( _filename.c_str(), ios::in );
	if( !f )
		throw new ErrorMsg( "Error opening HEX file for input!" );

	/*  */
	baseAddress = 0;
	start = size;
	end = 0;

	/* ںމ */
	f >> hexLine;							// 1sǂݍ
	while( !f.eof() )
	{
		Util.progress( "#" );				// i\is

		/* `ɏ]ںނ */
		parseRecord( hexLine, &rec );

		switch( rec.type )
		{
			case 0x00:		// ް ں?
				/* ް */
				if( baseAddress + rec.offset + rec.length > size )
					throw new ErrorMsg( "HEX file defines data outside buffer limits! "
										"Make sure file does not contain data outside device "
										"memory limits. "
										"Line from file was: (" + hexLine + ")." );

				for( dataPos = 0; dataPos < rec.length; dataPos++ )
				{
					data[ baseAddress + rec.offset + dataPos ] = rec.data[ dataPos ];
				}

				/* ޲Ă̎gXV */
				if( baseAddress + rec.offset < start )
					start = baseAddress + rec.offset;

				if( baseAddress + rec.offset + rec.length - 1 > end )
					end = baseAddress + rec.offset + rec.length - 1;

				break;

			case 0x02:		// g ڽ ں?
				baseAddress = (rec.data[0] << 8) | rec.data[1];
				baseAddress <<= 4;
				break;

			case 0x03:		// Jn ڽ ں?
				break;		// sJnڽ̉eȂ߁A𖳎

			case 0x04:		// gڽ ں?
				baseAddress = (rec.data[0] << 8) | rec.data[1];
				baseAddress <<= 16;
				break;

			case 0x05:		// Jnڽ ں?
				break;		// sJnڽ̉eȂ߁A𖳎

			case 0x01:		// ̧ ںނ̏I?
				f.close();
				Util.progress( "\r\n" );		// i\I
				return;

			default:
				throw new ErrorMsg( "Unsupported HEX record format! "
						"Line from file was: (" + hexLine + ")." );
		}

		f >> hexLine;						// sǂݍ
	}

	/* ŏIׂł͂܂B */
	throw new ErrorMsg( "Premature end of file encountered! Make sure file "
						"contains an EOF-record." );
}

void HEXFile::writeFile( const string & _filename )
{
	ofstream f;
	HEXRecord rec;							// ꎞں

	long baseAddress;						// ްΈʒu
	long offset;							// ڽ̵̾
	long dataPos;							// ް ںޓ̈ʒu

	enum
	{
		_first,
		_writing,
		_passed64k
	} status;								// ԏ݁A̎gB

	/* ٍ̧쐬 */
	f.open( _filename.c_str(), ios::out );
	if( !f )
		throw new ErrorMsg( "Error opening HEX file for output!" );

	/*  */
	status = _first;
	rec.data = new unsigned char[ 16 ];		// 16޲ ںނgp

	baseAddress = start & ~0xffff;			// 64K񂳂ꂽڽ
	offset = start & 0xffff;				// 񂳂ꂽڽ̵̾
	dataPos = 0;

	/* HEX̧قɍŏ̊ڽ ںނ */
	rec.length = 2;
	rec.offset = 0;
	rec.type = 0x02;
	rec.data[1] = 0x00;
	rec.data[0] = baseAddress >> 12;		// 4K߰ގwW^
	writeRecord( f, &rec );					// ̧قHEXںނ

	/* gp͈͂őSĂ޲Ă */
	do
	{
		/* ںޓްu */
		rec.data[ dataPos ] = data[ baseAddress + offset + dataPos ];
		dataPos++;

		/* ݂ް ںނ̏oKv */
		if( offset + dataPos >= 0x10000 ||	// 64KEɓB?
			dataPos >= 16 ||				// ް ںޖt?
			baseAddress + offset + dataPos > end )	// gp͈͂̍ŌɓB?
		{
			/* ݂ް ںނ */
			rec.length = dataPos;
			rec.offset = offset;
			rec.type = 0x00;				// ް ں

			Util.progress( "#" );			// i\is
			writeRecord( f, &rec );

			offset += dataPos;
			dataPos = 0;
		}

		/* 64KEʉ߂ */
		if( offset + dataPos >= 0x10000 )
		{
			/* ڽ ߲XV */
			offset -= 0x10000;
			baseAddress += 0x10000;

			/* HEX̧قɐVڽ ںނ */
			rec.length = 2;
			rec.offset = 0;
			rec.type = 0x02;
			rec.data[0] = baseAddress >> 12;	// 4K߰ގwW^
			rec.data[1] = 0x00;

			writeRecord( f, &rec );			// ̧قHEXںނ
		}
	} while( baseAddress + offset + dataPos <= end );

	/* EOFںނ */
	rec.length = 0;
	rec.offset = 0;
	rec.type = 0x01;

	writeRecord( f, &rec );

	f.close();
	Util.progress( "\r\n" );				// i\I
}

void HEXFile::setUsedRange( long _start, long _end )
{
	if( _start < 0 || _end >= size || _start > _end )
		throw new ErrorMsg( "Invalid range! Start must be 0 or larger, end must be "
							"inside allowed memory range." );

	start = _start;
	end = _end;
}

void HEXFile::clearAll( long value )
{
	for( long i = 0; i < size; i++ )
		data[i] = (unsigned char) (value & 0xff);
}

long HEXFile::getData( long address )
{
	if( address < 0 || address >= size )
		throw new ErrorMsg( "Address outside legal range!" );

	return data[ address ];
}

void HEXFile::setData( long address, long value )
{
	if( address < 0 || address >= size )
		throw new ErrorMsg( "Address outside legal range!" );

	data[ address ] = (unsigned char) (value & 0xff);
}

/* ̧ق̍Ō */
